/* Pinout of the SNES extension:
   BUFFER (8)           NOT USED
   B_8255 PORT A (8)    NOT USED
   B_8255 PORT B (8)    ADDR 11-12-10-13-9-14-8-15
   B_8255 PORT Cl (4)   ADDR 7-16-6-17
   B_8255 PORT Ch (4)   ADDR 5-18-4-19
   A_8255 PORT A (8)    ADDR 3-20-2-21-1-22-0-23
   A_8255 PORT B (8)    DATA 0-4-1-5-2-6-3-7
   A_8255 PORT Cl (4)   bit 0: /RD
						bit 1: /WR
						bit 2: /RAME
						bit 3: /POE
   A_8255 PORT Ch (4)   NOT USED
   GND,VCC... (8)
 */

#ifdef NAME
#undef NAME
#endif
#define NAME "SNES"

#define SNES_RD	 	0
#define SNES_WR		1
#define SNES_RAME	2
#define SNES_POE	3

#define SNES_IDLE		OUTPORT(A_8255_PC, 0xFF)
#define SNES_RD_ROM		OUTPORT(A_8255_PC, 0xFF-(0x01<<SNES_RD)-(0x01<<SNES_POE))
#define SNES_RD_LORAM	OUTPORT(A_8255_PC, 0xFF-(0x01<<SNES_RD)-(0x01<<SNES_POE))
#define SNES_RD_HIRAM	OUTPORT(A_8255_PC, 0xFF-(0x01<<SNES_RD))
#define SNES_WR_LORAM	OUTPORT(A_8255_PC, 0xFF-(0x01<<SNES_POE))
#define SNES_WR_HIRAM	OUTPORT(A_8255_PC, 0xFF)

#define SNES_LOROM	((snes_header[21]&0x0F)!=1)

unsigned char *snes_header = NULL;
unsigned long snes_sram_addr = 0;

INLINE int snes_read_byte(register unsigned long addr)
{
	register int i;

	OUTPORT(B_8255_PB,
		((addr>>11)&0x01) |
		((addr>>11)&0x02) |
		((addr>>8 )&0x04) |
		((addr>>10)&0x08) |
		((addr>>5 )&0x10) |
		((addr>>9 )&0x20) |
		((addr>>2 )&0x40) |
		((addr>>8 )&0x80));
	OUTPORT(B_8255_PC,
		((addr>>7 )&0x01) |
		((addr>>15)&0x02) |
		((addr>>4 )&0x04) |
		((addr>>14)&0x08) |
		((addr>>1 )&0x10) |
		((addr>>13)&0x20) |
		((addr<<2 )&0x40) |
		((addr>>12)&0x80));
	OUTPORT(A_8255_PA,
		((addr>>3 )&0x01) |
		((addr>>19)&0x02) |
		( addr     &0x04) |
		((addr>>18)&0x08) |
		((addr<<3 )&0x10) |
		((addr>>17)&0x20) |
		((addr<<6 )&0x40) |
		((addr>>16)&0x80));
	i = INPORT(A_8255_PB);
	return
		( i&0x01    ) |
		((i&0x02)<<3) |
		((i&0x04)>>1) |
		((i&0x08)<<2) |
		((i&0x10)>>2) |
		((i&0x20)<<1) |
		((i&0x40)>>3) |
		( i&0x80 );
}

INLINE void snes_write_byte(register int val, register unsigned long addr)
{
	SNES_IDLE;
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_out|PCh_in);
	/* Reconfiguring the 8255 may modify the PC register! */
	if(SNES_LOROM) {
		SNES_WR_LORAM;
	} else {
		SNES_WR_HIRAM;
	}
	OUTPORT(B_8255_PB,
		((addr>>11)&0x01) |
		((addr>>11)&0x02) |
		((addr>>8 )&0x04) |
		((addr>>10)&0x08) |
		((addr>>5 )&0x10) |
		((addr>>9 )&0x20) |
		((addr>>2 )&0x40) |
		((addr>>8 )&0x80));
	OUTPORT(B_8255_PC,
		((addr>>7 )&0x01) |
		((addr>>15)&0x02) |
		((addr>>4 )&0x04) |
		((addr>>14)&0x08) |
		((addr>>1 )&0x10) |
		((addr>>13)&0x20) |
		((addr<<2 )&0x40) |
		((addr>>12)&0x80));
	OUTPORT(A_8255_PA,
		((addr>>3 )&0x01) |
		((addr>>19)&0x02) |
		( addr     &0x04) |
		((addr>>18)&0x08) |
		((addr<<3 )&0x10) |
		((addr>>17)&0x20) |
		((addr<<6 )&0x40) |
		((addr>>16)&0x80));
	OUTPORT(A_8255_PB,
		( val&0x01    ) |
		((val>>3)&0x02) |
		((val<<1)&0x04) |
		((val>>2)&0x08) |
		((val<<2)&0x10) |
		((val>>1)&0x20) |
		((val<<3)&0x40) |
		( val&0x80 ));
	OUTPORT(A_8255_PROG, CLR_PC(SNES_WR));
	OUTPORT(A_8255_PROG, SET_PC(SNES_WR));
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_in|PCl_out|PCh_in);
}

void snes_read_header()
{
	int i;

	SNES_RD_ROM;
	if(snes_header != NULL)
		return;

	snes_header = (unsigned char *)malloc(0x40 * sizeof(unsigned char));

	for(i = 0; i < 0x40; i++) {
		snes_header[i] = snes_read_byte(0xFFC0 + i);
	}
}

void snes_init()
{
	check_extension(NAME);

	/* Configure 8255's */
	/* B:
			port A: INPUT     NOT USED
			port B: OUTPUT    ADDR 11-12-10-13-9-14-8-15
			port Cl: OUTPUT   ADDR 7-16-6-17
			port Ch: OUTPUT   ADDR 5-18-4-19
	   A:
			port A: OUTPUT    ADDR 3-20-2-21-1-22-0-23
			port B: INPUT     DATA 0-4-1-5-2-6-3-7
			port Cl: OUTPUT   bit 0: /RD
							  bit 1: /WR
							  bit 2: /RAME
							  bit 3: /POE
			port Ch: INPUT    NOT USED

	*/
	OUTPORT(B_8255_PROG, 0x80|PA_in|PB_out|PCl_out|PCh_out);
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_in|PCl_out|PCh_in);
	SNES_IDLE;
}

void snes_analyze()
{
	char name[22];
	char *s;
	unsigned int checksum;

	snes_read_header();

	strncpy(name, &snes_header[0], 21);
	name[21] = 0;
	printf("Game name: %s\n", name);
	switch((snes_header[21]&0xF0)>>4) {
	case 0:
		s = "SlowROM";
		break;
	case 3:
		s = "FastROM";
		break;
	default:
		s = "Unknown";
		break;
	}
	printf("ROM speed: %s\n", s);
	switch(snes_header[21]&0x0F) {
	case 0:
		s = "LoROM (32 kB banks)";
		break;
	case 1:
		s = "HiROM (64 kB banks)";
		break;
	default:
		s = "Unknown";
		break;
	}
	printf("Bank size: %s\n", s);
	switch(snes_header[22]) {
	case 0:
		s = "ROM only";
		break;
	case 1:
		s = "ROM and RAM";
		break;
	case 2:
		s = "ROM and Save RAM";
		break;
	case 3:
		s = "ROM and DSP1 chip";
		break;
	case 4:
		s = "ROM, RAM and DSP1 chip";
		break;
	case 5:
		s = "ROM, Save RAM and DSP1 chip";
		break;
	case 19:
		s = "ROM and Super FX chip";
		break;
	case 227:
		s = "ROM, RAM and GameBoy data";
		break;
	case 246:
		s = "ROM and DSP2 chip";
		break;
	default:
		s = "Unknown";
		break;
	}
	printf("ROM type: %d(%s)\n", snes_header[22], s);
	printf("ROM size: %d(%d kB)\n", snes_header[23], 1 << snes_header[23]);
	printf("SRAM size: %d(%d kB)\n", snes_header[24], snes_header[24] == 0 ? 0 : 1 << snes_header[24]);
	if(snes_header[24] > 0 && snes_sram_addr == 0) {
		if(SNES_LOROM)
			snes_sram_addr = 0x700000;
		else
			snes_sram_addr = 0x206000;
		printf("SRAM addr: %06lX\n", snes_sram_addr);
	}
	checksum = snes_header[30]+(snes_header[31]<<8);
	printf("Checksum: %02X\n", checksum);
}

void snes_save(FILE *fp)
{
	unsigned long addr, min, max, sizekB, chunk;
	unsigned long *checksums;
	unsigned int checksum, chk;
	int i, nbchunk;

	snes_analyze();

	sizekB = 1 << snes_header[23];

	/* One chunk is 4Mbit = 512kB */
	nbchunk = sizekB / 512;
	checksums = (long *)calloc(nbchunk, sizeof(long));

	SNES_RD_ROM;
	for(chunk = 0; chunk < nbchunk; chunk++) {
		printf("Reading chunk %ld\n", chunk);
		min = chunk << 19;
		max = (chunk + 1) << 19;
		if(SNES_LOROM) {
			min <<= 1;
			max <<= 1;
		}
		for(addr = min; addr < max; addr++) {
			if(addr % 0x1000 == 0) {
				putchar('.');
				fflush(stdout);
			}
			i = snes_read_byte(addr);
			checksums[chunk] += i;
			fputc(i, fp);
			if(SNES_LOROM && (addr % 0x8000 == 0x7FFF))
				addr += 0x8000;
		}
		putchar('\n');
	}

	chk = 0;
	for(i = 0; i < nbchunk; i++)
		chk += checksums[i];
	chk &= 0xFFFF;
	checksum = snes_header[30]+(snes_header[31]<<8);
	if(chk == checksum)
		printf("Checksum OK\n");
	else
		printf("Warning: checksum is wrong (waiting for %X, got %X)\n", checksum, chk);
	free(checksums);
}

void snes_backup(FILE *fp)
{
	unsigned long addr, max, sizekB;
	int i;

	snes_analyze();

	if(snes_header[24] == 0) {
		printf("No save-RAM\n");
		return;
	}
	sizekB = 1 << snes_header[24];

	if(SNES_LOROM) {
		SNES_RD_LORAM;
	} else {
		SNES_RD_HIRAM;
	}
	printf("Reading SRAM\n");
	max = snes_sram_addr+(sizekB<<10);
	for(addr = snes_sram_addr; addr < max; addr++) {
		if(addr % 0x1000 == 0) {
			putchar('.');
			fflush(stdout);
		}
		i = snes_read_byte(addr);
		fputc(i, fp);
	}
	putchar('\n');
}

void snes_restore(FILE *fp)
{
	unsigned long addr, max, sizekB;
	unsigned char val;

	snes_analyze();

	if(snes_header[24] == 0) {
		printf("No save-RAM\n");
		return;
	}
	sizekB = 1 << snes_header[24];

	printf("Restoring SRAM\n");
	max = snes_sram_addr+(sizekB<<10);
	for(addr = snes_sram_addr; addr < max; addr++) {
		if(feof(fp)) {
			printf("Unexpected EOF\n");
			exit(1);
		}
		if(addr % 0x1000 == 0) {
			putchar('.');
			fflush(stdout);
		}
		val = fgetc(fp);
		snes_write_byte(val, addr);
	}
	putchar('\n');
}

void snes_clear()
{
	unsigned long addr, max, sizekB;

	snes_analyze();

	if(snes_header[24] == 0) {
		printf("No save-RAM\n");
		return;
	}
	sizekB = 1 << snes_header[24];

	printf("Clearing SRAM\n");
	max = snes_sram_addr+(sizekB<<10);
	for(addr = snes_sram_addr; addr < max; addr++) {
		if(addr % 0x1000 == 0) {
			putchar('.');
			fflush(stdout);
		}
		snes_write_byte(0x00, addr);
	}
	putchar('\n');
}

void snes_test()
{
	unsigned long addr, max, sizekB;
	int i, test_val;

	snes_analyze();

	if(snes_header[24] == 0) {
		printf("No save-RAM\n");
		return;
	}
	sizekB = 1 << snes_header[24];

	printf("Testing SRAM\n");
	max = snes_sram_addr+(sizekB<<10);

	for(i = 0; test_patterns[i] >= 0; i++) {
		test_val = test_patterns[i];
		printf("Writing 0x%02X's\n", test_val);
		for(addr = snes_sram_addr; addr < max; addr++) {
			if(addr % 0x1000 == 0) {
				putchar('.');
				fflush(stdout);
			}
			snes_write_byte(test_val, addr);
		}
		putchar('\n');
		if(SNES_LOROM) {
			SNES_RD_LORAM;
		} else {
			SNES_RD_HIRAM;
		}
		printf("Verifying 0x%02X's\n", test_val);
		for(addr = snes_sram_addr; addr < max; addr++) {
			if(addr % 0x1000 == 0) {
				putchar('.');
				fflush(stdout);
			}
			if(snes_read_byte(addr) != test_val) {
				fprintf(stderr, "\nError at addr 0x%06lX\n", addr);
				exit(-1);
			}
		}
		putchar('\n');
	}
	printf("OK\n");
}

void snes_debug()
{
	char s[32];
	int val;
	unsigned long addr;

	while(1) {
		printf("\n[Command,'?']-> ");
		fflush(stdout);
		fflush(stdin);

		fgets(s, 31, stdin);
		for(val = 0; s[val] >= ' '; val++)
			s[val] = toupper(s[val]);
		s[val]='\0';

		switch(s[0]) {
			case 'H':
			case '?':
				puts("C<val>  : Set command bits (RD=1,WR=2,RAME=4,POE=8) (hex)");
				puts("A<addr> : Set address (hex)");
				puts("Q       : Exit");
				break;
			case 'C':
				sscanf(s+1, "%X", &val);
				val |= 0xF0;
				printf("Command = %X\n", val);
				OUTPORT(A_8255_PC, val);
				break;
			case 'A':
				sscanf(s+1, "%lX", &addr);
				printf("Addr    = %lX\n", addr);
				OUTPORT(B_8255_PB,
					((addr>>11)&0x01) |
					((addr>>11)&0x02) |
					((addr>>8 )&0x04) |
					((addr>>10)&0x08) |
					((addr>>5 )&0x10) |
					((addr>>9 )&0x20) |
					((addr>>2 )&0x40) |
					((addr>>8 )&0x80));
				OUTPORT(B_8255_PC,
					((addr>>7 )&0x01) |
					((addr>>15)&0x02) |
					((addr>>4 )&0x04) |
					((addr>>14)&0x08) |
					((addr>>1 )&0x10) |
					((addr>>13)&0x20) |
					((addr<<2 )&0x40) |
					((addr>>12)&0x80));
				OUTPORT(A_8255_PA,
					((addr>>3 )&0x01) |
					((addr>>19)&0x02) |
					( addr     &0x04) |
					((addr>>18)&0x08) |
					((addr<<3 )&0x10) |
					((addr>>17)&0x20) |
					((addr<<6 )&0x40) |
					((addr>>16)&0x80));
				break;
			case 'Q':
				exit(0);
		}
	}
}

Extension snes_ext = {
	NAME,
	snes_init,
	NULL,		/* no_flag */
	snes_analyze,
	snes_save,
	snes_backup,
	snes_restore,
	snes_clear,
	snes_test,
	snes_debug,
	/*	NULL,*/		/* debug */
	NULL,		/* extra */
	NULL		/* usage */
};
