/* Pinout of the VB extension:
   BUFFER (8)			NOT USED
   B_8255 PORT A (8)    ADDR NC-NC-NC-18-20-17-19-7
   B_8255 PORT B (8)    ADDR 8-6-9-5-10-4-11-3
   B_8255 PORT Cl (4)   ADDR 12-2-13-1
   B_8255 PORT Ch (4)   ADDR 14-0-15-16
   A_8255 PORT A (8)    DATA 15-0-7-8-14-1-6-9
   A_8255 PORT B (8)    DATA 13-2-5-10-12-3-4-11

   A_8255 PORT Cl (4)   bit 0: /OE
				bit 1: /CE
				bit 2: /WE		//Flash Rom
				bit 3: /RESET	//Flash Rom

   A_8255 PORT Ch (4)   bit 4: /CS1		//Ram
				bit 5:  CS2
				bit 6: /WE
				bit 7:  NOT USED
   GND,VCC... (8)
 */
//===================================
// Defines
//===================================

#ifdef NAME
#undef NAME
#endif
#define NAME "VB"

#define BYTE	unsigned char //8bit
#define HWORD	unsigned int  //16 bit
#define WORD	unsigned long //32 bit

#define VB_OE    0 //shared between ram and rom
#define VB_CE    1

#define VB_WE    2 //flash
#define VB_RESET 3 //flash

#define VB_RAM_CS1 4
#define VB_RAM_CS2 5
#define VB_RAM_WE  6

#define bit(x)               (0x01<<x)
#define ctrl_set(x)          OUTPORT(A_8255_PC, 0xFF-(x))

//#define VB_IDLE              bit(VB_RAM_CS2)
//#define VB_READ_ROM          bit(VB_RAM_CS2)|bit(VB_CE)
//#define VB_WRITE_ROM         bit(VB_RAM_CS2)|bit(VB_CE)
//#define VB_READ_SRAM         bit(VB_RAM_CS1)
//#define VB_WRITE_SRAM        bit(VB_RAM_CS1)|bit(VB_RAM_WE)

BYTE vb_header[0x20];

//===================================
// Delay Fn
//===================================

int dt_delay(int ms){
	int i,j,k;
	for(k=0;k<ms;k++) {
		for(i=0;i<5;i++) {
			j = j*i; //100
		}
	}
//	delay(ms/20);
	return 1;
}

//===================================
// Helper Fn
//===================================

//write an address to the bus
inline void writeaddr(WORD addr) {
	OUTPORT(B_8255_PA,
		(((0   >> 0)&1)<<0) |  //NC
		(((0   >> 0)&1)<<1) |  //NC
		(((0   >> 0)&1)<<2) |  //NC
		(((addr>>18)&1)<<3) |  //18
		(((addr>>20)&1)<<4) |  //20
		(((addr>>17)&1)<<5) |  //17
		(((addr>>19)&1)<<6) |  //19
		(((addr>> 7)&1)<<7));  // 7
	OUTPORT(B_8255_PB,
		(((addr>> 8)&1)<<0) |  // 8
		(((addr>> 6)&1)<<1) |  // 6
		(((addr>> 9)&1)<<2) |  // 9
		(((addr>> 5)&1)<<3) |  // 5
		(((addr>>10)&1)<<4) |  //10
		(((addr>> 4)&1)<<5) |  // 4
		(((addr>>11)&1)<<6) |  //11
		(((addr>> 3)&1)<<7));  // 3
	OUTPORT(B_8255_PC,
		(((addr>>12)&1)<<0) |  //12
		(((addr>> 2)&1)<<1) |  // 2
		(((addr>>13)&1)<<2) |  //13
		(((addr>> 1)&1)<<3) |  // 1
		(((addr>>14)&1)<<4) |  //14
		(((addr>> 0)&1)<<5) |  // 0
		(((addr>>15)&1)<<6) |  //15
		(((addr>>16)&1)<<7));  //16
}

//convert the databus data back into a proper hword
inline HWORD data2hword(HWORD data) {
	return
		((data&0x0001)<<15) |
		((data&0x0002)>>1 ) |
		((data&0x0004)<<5 ) |
		((data&0x0008)<<5 ) |
		((data&0x0010)<<10) |
		((data&0x0020)>>4 ) |
		( data&0x0040     ) |
		((data&0x0080)<<2 ) |
		((data&0x0100)<<5 ) |
		((data&0x0200)>>7 ) |
		((data&0x0400)>>5 ) |
		((data&0x0800)>>1 ) |
		( data&0x1000     ) |
		((data&0x2000)>>10) |
		((data&0x4000)>>10) |
		((data&0x8000)>>4 );
}

//convert a word of data to the databus format
inline HWORD hword2data(HWORD hword) {
	return
		((hword&0x8000)>>15) |
		((hword&0x0001)<<1 ) |
		((hword&0x0080)>>5 ) |
		((hword&0x0100)>>5 ) |
		((hword&0x0400)>>10) |
		((hword&0x0002)<<4 ) |
		( hword&0x0040     ) |
		((hword&0x0200)>>2 ) |
		((hword&0x2000)>>5 ) |
		((hword&0x0004)<<7 ) |
		((hword&0x0020)<<5 ) |
		((hword&0x0400)<<1 ) |
		( hword&0x1000     ) |
		((hword&0x0008)<<10) |
		((hword&0x0010)<<10) |
		((hword&0x0800)<<4 );
}

//===================================
// Read/Write ROM
//===================================

INLINE HWORD vb_read_word(WORD addr)
{
	HWORD i,j,k;

	writeaddr(addr);

	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE)|bit(VB_OE));
	dt_delay(100);

	i = INPORT(A_8255_PA) | (INPORT(A_8255_PB) << 8);
	j = INPORT(A_8255_PA) | (INPORT(A_8255_PB) << 8);
	k = INPORT(A_8255_PA) | (INPORT(A_8255_PB) << 8);
	if((i!=j)||(i!=k)||(j!=k)) 
		printf("error reading addr 0x%04X: 0x%04X, 0x%04X, 0x%04X\n",addr,i,j,k);

	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE)); // Reset the port
	dt_delay(20);

	return data2hword(i);
}

INLINE HWORD vb_write_word(HWORD data, WORD addr)
{
	HWORD i;

	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE));

	writeaddr(addr);

	dt_delay(1);

	i = hword2data(data);

	OUTPORT( A_8255_PA,  i      &0xFF );
	OUTPORT( A_8255_PB, (i >> 8)&0xFF );
	dt_delay(1);
	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE)|bit(VB_WE));
	dt_delay(1);
	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE)|bit(VB_WE)|bit(VB_OE));
	dt_delay(1);
	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE));
	return 1;
}

//===================================
// Read/Write SRAM
//===================================

INLINE BYTE vb_read_sram(WORD addr)
{
	HWORD i,j,k;

	writeaddr(addr);

	dt_delay(5);

	ctrl_set(bit(VB_RAM_CS1)|bit(VB_OE));
	dt_delay(100);
	i = INPORT(A_8255_PA) | (INPORT(A_8255_PB) << 8);
	j = INPORT(A_8255_PA) | (INPORT(A_8255_PB) << 8);
	k = INPORT(A_8255_PA) | (INPORT(A_8255_PB) << 8);

	i =data2hword(i)&0x00FF;
	j =data2hword(j)&0x00FF;
	k =data2hword(k)&0x00FF;

	if((i!=j)||(i!=k)||(j!=k)) 
		printf("error reading addr 0x%04X: 0x%02X, 0x%02X, 0x%02X\n",addr,i,j,k);

	ctrl_set(bit(VB_RAM_CS1)); // Reset the port
	dt_delay(40);

	//return the lower 8 bits
	return i;
}

INLINE HWORD vb_write_sram(HWORD data, WORD addr)
{
	HWORD i;

	//Set Up Address
	writeaddr(addr);

	ctrl_set(bit(VB_RAM_CS1));
	dt_delay(200);

	//Set Up Data
	i = hword2data(data);
	OUTPORT(A_8255_PA,  i      &0xFF);
	OUTPORT(A_8255_PB, (i >> 8)&0xFF);
	dt_delay(200);

	ctrl_set(bit(VB_RAM_CS1)|bit(VB_RAM_WE));
	dt_delay(200);
	//dt_delay(2000);
	ctrl_set(bit(VB_RAM_CS2));

	return 1;
}

//===================================
// Public Fn's
//===================================

int vb_get_rom_size() {
	HWORD val = 32;
	HWORD i = 23;
	WORD addr = 0x13;
	WORD off = 0x40000;

	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE)); // Reset the port

	val = vb_read_word(addr);
	if((val == 0) || (val == 0xff)) return 0x100000;
	i = vb_read_word(addr + off);
	if(val == i) return off;
	off = 0x80000;
	i = vb_read_word(addr + off);
	if(val == i) return off;
	off = 0x100000;

	ctrl_set(bit(VB_RAM_CS2));

	return off;
}

void vb_read_header()
{
	HWORD val, i;
	WORD addr;

	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE)); // Reset the port


	i = 0;
	for(addr = 0xFFFFFEF0; addr < 0xFFFFFF00; addr++) {
		val = vb_read_word(addr);
		vb_header[i++] = (BYTE)val;
		vb_header[i++] = (BYTE)(val>>8);
	}

	ctrl_set(bit(VB_RAM_CS2));
}

void vb_init()
{
	check_extension(NAME);

	/* Configure 8255's */
	/* B:
			port A: OUTPUT    ADDR NC-NC-NC-18-20-17-19-7
			port B: OUTPUT    ADDR 8-6-9-5-10-4-11-3
			port Cl: OUTPUT   ADDR 12-2-13-1
			port Ch: OUTPUT   ADDR 14-0-15-16
	   A:
			port A: INPUT     DATA 15-0-7-8-14-1-6-9
			port B: INPUT     DATA 13-2-5-10-12-3-4-11
			port cl: OUTPUT	bit 0: /OE
						bit 1: /CE
						bit 2: /WE		//Flash Rom
						bit 3: /RESET	//Flash ROm
			port ch: OUTPUT	bit 4: /CS1		//Ram
						bit 5:  CS2
						bit 6: /WE
						bit 7:  NOT USED
	*/
	OUTPORT(B_8255_PROG, 0x80|PA_out|PB_out|PCl_out|PCh_out);
	OUTPORT(A_8255_PROG, 0x80|PA_in|PB_in|PCl_out|PCh_out);
	ctrl_set(bit(VB_RAM_CS2));
}

void vb_analyze()
{
	char vb_name[21];
	char vb_manufacture[3];
	char vb_gcode[5];
	WORD rom_size = 0;

	vb_read_header();

	strncpy(vb_name, &vb_header[0], 20);
	vb_name[20] = 0;

	strncpy(vb_manufacture, &vb_header[25], 2);
	vb_manufacture[2] = 0;

	strncpy(vb_gcode, &vb_header[27], 4);
	vb_gcode[4] = 0;

	printf("Game name: %s\n", vb_name);
	printf("Game manufacture: %s\n", vb_manufacture);
	printf("Game code: %s\n", vb_gcode);

	//Fix me up...
	rom_size = vb_get_rom_size();
	rom_size = (rom_size*2)/1024;

	printf("Rom Size: %d KByte\n", rom_size); //08l
}

//Save the rom to a file
void vb_save_rom(FILE *fp)
{
	HWORD val;
	WORD addr;
	HWORD checksum;


	printf("Saving ROM\n");
	vb_analyze();

	checksum = 0;
	ctrl_set(bit(VB_RAM_CS2)|bit(VB_CE)); // Reset the port

// Uncomment if romes dont dump right!!!
//	for(addr = 0x000000; addr < 0x100000; addr++) {
	for(addr = 0x000000; addr < vb_get_rom_size(); addr++) {
		if(addr % 0x1000 == 0) {
			putchar('.');
			fflush(stdout);
		}
		val = vb_read_word(addr);
		checksum += val;
		fputc((val&0xFF), fp);
		fputc(((val>>8)&0xFF), fp);
	}
	putchar('\n');
	printf("Checksum = 0x%04X\n", checksum);

	ctrl_set(bit(VB_RAM_CS2));
}


//Save the ram to a file
void vb_save_sram(FILE *fp)
{
	BYTE cVal;
	WORD addr;
	HWORD checksum;

	printf("Saving SRAM\n");
	vb_analyze();

	checksum = 0;

	ctrl_set(bit(VB_RAM_CS1)); // Reset the port

	//8KB
	for(addr = 0x000000; addr < 0x2000; addr++) {
		if(addr % 0x500 == 0) {
			putchar('.');
			fflush(stdout);
		}
		cVal = vb_read_sram(addr);
		checksum += cVal;
		fputc((cVal&0xFF), fp);
	}
	putchar('\n');
	printf("Checksum = 0x%04X\n", checksum);

	ctrl_set(bit(VB_RAM_CS2));
}

//Write SRAM from file
void vb_restore_sram(FILE *fp)
{
	BYTE cVal;
	WORD addr=0;

	printf("Writing SRAM\n");
	vb_analyze();

	//set to write mode
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_out|PCh_out);
	ctrl_set(bit(VB_RAM_CS2)); // Reset the port

	//write a bogus value to initiate the write cycle
	vb_write_sram(0x11, 0x00);

	addr = 0x00000000;
	while((!feof(fp))&&(addr<0x2000)) {
		if(addr % 0x500 == 0) {
			putchar('.');
			fflush(stdout);
		}

		cVal = fgetc(fp)&0x00FF;
		vb_write_sram(cVal, addr);
		addr++;
	}
	putchar('\n');

	//reset to read mode
	OUTPORT(A_8255_PROG, 0x80|PA_in|PB_in|PCl_out|PCh_out);
	ctrl_set(bit(VB_RAM_CS2));
}

//Try and debug the interface
void vb_debug()
{
	HWORD val;
	WORD addr;

	printf("Debug Connection\n");

//===Initialize===

	//set to write mode
	OUTPORT(A_8255_PROG, 0x80|PA_out|PB_out|PCl_out|PCh_out);
	//reset addres lines to zero
	writeaddr(0x00000000);

	//clear the data lines
	val = hword2data(0x00000000);
	OUTPORT( A_8255_PA,  val      &0xFF );
	OUTPORT( A_8255_PB, (val >> 8)&0xFF );
	
	//clear the controll lines
	ctrl_set(bit(VB_OE)|bit(VB_CE)|bit(VB_WE)|bit(VB_RESET)|bit(VB_RAM_CS1)|bit(VB_RAM_CS2)|bit(VB_RAM_WE));

	printf("\n - Initial state, all lines zero <enter>\n");
	getchar();

//===Test controll lines===

	printf("\n - Testing controll lines\n");

	ctrl_set(0); //Ctrl Lines high
	printf("    Ctrl Lines, all lines high <enter>\n");
	getchar();

	ctrl_set(bit(VB_OE));
	printf("    Ctrl Lines, VB_OE low <enter>\n");
	getchar();

	ctrl_set(bit(VB_CE));
	printf("    Ctrl Lines, VB_CE low <enter>\n");
	getchar();

	ctrl_set(bit(VB_WE));
	printf("    Ctrl Lines, VB_WE low <enter>\n");
	getchar();

	ctrl_set(bit(VB_RESET));
	printf("    Ctrl Lines, VB_RESET low <enter>\n");
	getchar();

	ctrl_set(bit(VB_RAM_CS1));
	printf("    Ctrl Lines, VB_RAM_CS1 low <enter>\n");
	getchar();

	ctrl_set(bit(VB_RAM_CS2));
	printf("    Ctrl Lines, VB_RAM_CS2 low <enter>\n");
	getchar();

	ctrl_set(bit(VB_RAM_WE));
	printf("    Ctrl Lines, VB_RAM_WE low <enter>\n");
	getchar();

//===Test data lines===

	ctrl_set(bit(VB_OE)|bit(VB_CE)|bit(VB_WE)|bit(VB_RESET)|bit(VB_RAM_CS1)|bit(VB_RAM_CS2)|bit(VB_RAM_WE));
	printf("\n - Testing data lines\n");

	val = hword2data(0xFFFF);
	OUTPORT( A_8255_PA,  val      &0xFF );
	OUTPORT( A_8255_PB, (val >> 8)&0xFF );

	printf("    Data Lines High <enter>\n");
	getchar();

	val = hword2data(0x0000);
	OUTPORT( A_8255_PA,  val      &0xFF );
	OUTPORT( A_8255_PB, (val >> 8)&0xFF );

	printf("    Data Lines Low <enter>\n");
	getchar();

//===Test Address lines===

	ctrl_set(bit(VB_OE)|bit(VB_CE)|bit(VB_WE)|bit(VB_RESET)|bit(VB_RAM_CS1)|bit(VB_RAM_CS2)|bit(VB_RAM_WE));
	printf("\n - Testing address lines\n");

	writeaddr(0x00FFFFFF);
	printf("    Address Lines High <enter>\n");
	getchar();

	writeaddr(0x00000000);
	printf("    Address Lines Low <enter>\n");
	getchar();

	for(addr = 0x000001; addr <= 0x100000 ; addr <<= 1) {
		writeaddr(addr);
		printf("Addr: 0x%06lX <enter>\n", addr);
		getchar();
	}

	//Set to read mode
	OUTPORT(A_8255_PROG, 0x80|PA_in|PB_in|PCl_out|PCh_out);

	//reset to Idle state
	ctrl_set(bit(VB_RAM_CS2));
}

//===================================
// Public definition
//===================================

Extension vb_ext = {
	NAME,
	vb_init,
	NULL,			/* no_flag */
	vb_analyze ,
	vb_save_rom,	/* backup rom */
	vb_save_sram,	/* backup ram */
	vb_restore_sram,  /* restore ram */
	NULL,			/* clear ram */
	NULL,			/* test */
	vb_debug,
	NULL,			/* extra */
	NULL			/* usage */
};
